# -*- coding: utf-8 -*-
"""Module with interfaces for using PyTorch in pailab.

This module provides some useful methods and code snippets to integrate PyTorch projects into pailab.
The main entry point is the method `add_model` that setups all methods and objects needed to train the individual 
PyTorch model. 

Note:
    Currently, pailab supports only numpy data which is connected via a DataSet object to the training and test functions 
    in the repository. Pailab's DataSet object has nothing to do with PyTorch's DataSet object and to use all functionality out of the box 
    one has to convert the given PyTorch DataSet object into a numpy array (this module provides the method `convert_to_numpy` to do this for simple DataSets).
    It may not be difficult to integrate PyTorch DataSets (e.g. using a specialization of class NumpyStore) in a more clever and adapted way which will be subject 
    of future work.
"""
import numpy as np
from collections import OrderedDict

import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader

from pailab.ml_repo.repo_objects import repo_object_init

from pailab.ml_repo.repo_objects import RepoInfoKey
from pailab.ml_repo.repo import MLObjectType
import logging
logger = logging.getLogger(__name__)


def _get_object_from_classname(classname, data):
    """ Returns an object instance for given classname and data dictionary.

    Args:
        classname (str): Full classname as string including the modules, e.g. repo.Y if class Y is defined in module repo.
        data (dict): dictionary of data used to initialize the object instance.

    Returns:
        [type] -- Instance object of class.
    """

    parts = classname.split('.')
    module = ".".join(parts[:-1])
    m = __import__(module)
    for comp in parts[1:]:
        m = getattr(m, comp)
    result = m(**data)  # ._create_from_dict(**data)
    return result


class PytorchModelWrapper:
    @repo_object_init(['state_dict'])
    def __init__(self, pytorch_model, param):
        self.classname = pytorch_model.__class__.__module__ + \
            '.' + pytorch_model.__class__.__name__
        self.state_dict_order = []
        self.state_dict = {}
        self.model_param = param
        for k, v in pytorch_model.state_dict().items():
            self.state_dict_order.append(k)
            self.state_dict[k] = v.numpy()

    def numpy_to_dict(self):
        return self.state_dict

    def numpy_from_dict(self, numpy_dict):
        self.state_dict = numpy_dict

    def get_model(self):
        state_dict = OrderedDict()
        for k in self.state_dict_order:
            state_dict[k] = torch.from_numpy(self.state_dict[k])
        model = _get_object_from_classname(self.classname, self.model_param)
        model.load_state_dict(state_dict)
        # model.eval()
        return model


class _PytorchDataset(Dataset):
    def __init__(self, data):
        self.data = torch.from_numpy(data.x_data).float()
        self.target = None
        if hasattr(data, 'y_data'):
            self.target = torch.from_numpy(data.y_data).float()
        self.transform = None

    def __getitem__(self, index):
        x = self.data[index]
        if self.transform:
            x = self.transform(x)
        if self.target is not None:
            y = self.target[index]
            return x, y
        else:
            return x

    def __len__(self):
        return len(self.data)


class PytorchTrainingParameter:
    @repo_object_init()
    def __init__(self, batch_size, epochs, loss='MSE', optimizer='Adam', optim_param={}):
        self.batch_size = batch_size
        self.epochs = epochs
        self.loss = loss
        self.loss = 'MSE'
        self.optimizer = optimizer
        # create all default parameter for the chosen optimizer
        optimizer_ = _get_object_from_classname(
            'torch.optim.' + optimizer, {'params': (torch.empty(1, 1, dtype=torch.float),)})
        optimizer_.defaults.update(optim_param)
        self.optim_param = optimizer_.defaults


class PytorchModelParameter:
    @repo_object_init()
    def __init__(self, pytorch_model, **kwargs):
        self.model_classname = pytorch_model.__class__.__module__ + \
            '.' + pytorch_model.__class__.__name__
        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_param(self):
        model_param = {name: attr for name, attr in self.__dict__.items()
                       if not name.startswith("__")
                       and not callable(attr)
                       and not type(attr) is staticmethod
                       and not name == 'model_classname'
                       and not name == 'repo_info'
                       }
        return model_param

    def create_model(self):
        model_param = self.get_param()
        model = _get_object_from_classname(self.model_classname, model_param)
        return model


def eval_pytorch(model: PytorchModelWrapper, data, num_workers=0):
    d = _PytorchDataset(data)
    loader = torch.utils.data.DataLoader(
        d, batch_size=1, num_workers=num_workers)
    m = model.get_model()
    m.eval()
    # determine number of data
    n_data = len(loader)
    result = None
    is_tuple = False
    for x in loader:
        if isinstance(x, list):
            output = m(x[0])
        else:
            output = m(x)
        if isinstance(output, list):
            result = np.empty((n_data, ) + tuple(output[0].shape[1:]))
            is_tuple = True
        else:
            result = np.empty((n_data, ) + tuple(output.shape[1:]))
        break
    if is_tuple:
        for i, x in enumerate(loader):
            if isinstance(x, list):
                output = m(x[0])
            else:
                output = m(x)
            result[i] = output.data
    else:
        for i, x in enumerate(loader):
            if isinstance(x, list):
                output = m(x[0])
            else:
                output = m(x)
            result[i] = output.data  # .detach.numpy()
    return result


def train_pytorch(model_param: PytorchModelParameter, train_param: PytorchTrainingParameter, data):
    if train_param.loss == 'MSE':
        criterion = nn.MSELoss()
    else:
        raise Exception('Unknown loss function ' + train_param.loss)

    # specify loss function
    logger.debug('Setting up data loader.')
    train_data = _PytorchDataset(data)
    train_loader = torch.utils.data.DataLoader(
        train_data, batch_size=train_param.batch_size, num_workers=0)
    logger.debug('Finished setting up data loader.')
    # create model
    logger.debug('Setting up model.')
    model = model_param.create_model()
    model.train()
    logger.debug('Finished setting up model.')

    optim_args = {'params': model.parameters()}
    optim_args.update(train_param.optim_param)
    optimizer = _get_object_from_classname('torch.optim.'+train_param.optimizer,
                                           optim_args)
    logger.info('Start training with ' + str(train_param.epochs) + ' epochs.')
    train_loss = 0.0
    for epoch in range(1, train_param.epochs+1):
        train_loss = 0.0
        for data in train_loader:
            #logger.debug('Start training')
            if isinstance(data, list):
                x, target = data
            else:
                x = data
                target = data
            optimizer.zero_grad()
            outputs = model(x)
            # calculate the loss
            loss = criterion(outputs, target)
            loss.backward()
            optimizer.step()
            train_loss += loss.item()*target.size(0)  # funktioniert nur für mean-values

        # print avg training statistics
        train_loss = train_loss/len(train_loader)
        logger.debug('Epoch: {} \tTraining Loss: {:.6f}'.format(
            epoch,
            train_loss
        ))

    logger.info('Finished training with train loss ' + str(train_loss))
    result = PytorchModelWrapper(model, model_param.get_param(), repo_info={})
    return result


def add_model(repo, pytorch_model, model_param, model_name,
              batch_size, epochs, loss='MSE', optimizer='Adam', optim_param={}):
    """Add a PyTorch model.

    This method adds all relevant objects to train the PyTorch model: The training and evaluation functions as well as training- and model parameter.
    The objects created for the training and model parameter are instances of the internally defined  PytorchTrainingParameter and PytorchModelParameter
    classes.

    Args:
        repo ([type]): [description]
        pytorch_model ([type]): [description]
        model_param ([type]): [description]
        model_name ([type]): [description]
        batch_size ([type]): [description]
        epochs ([type]): [description]
        loss (str, optional): [description]. Defaults to 'MSE'.
        optimizer (str, optional): [description]. Defaults to 'Adam'.
        optim_param (dict, optional): [description]. Defaults to {}.
    """
    repo.add_eval_function(eval_pytorch,
                           repo_name='eval_pytorch')
    repo.add_training_function(
        train_pytorch, repo_name='train_pytorch')

    train_param = PytorchTrainingParameter(
        batch_size, epochs, loss, optimizer, optim_param,  repo_info={RepoInfoKey.NAME.value: model_name + '/training_param',
                                                                      RepoInfoKey.CATEGORY: MLObjectType.TRAINING_PARAM})
    model_param = PytorchModelParameter(
        pytorch_model, **model_param, repo_info={RepoInfoKey.NAME.value: model_name + '/model_param',
                                                 RepoInfoKey.CATEGORY: MLObjectType.MODEL_PARAM})

    repo.add([train_param, model_param])
    repo.add_model(model_name, model_eval='eval_pytorch',
                   model_training='train_pytorch', model_param=model_name + '/model_param', training_param=model_name + '/training_param')


def convert_to_numpy(pt_data, tuple_index=0):
    """Convert a PyTorch DataSet into a numpy array.

    Args:
        pt_data (torch.utils.data.DataSet): The DataSet to be converted to numpy.
        tuple_index (int): If each item of the dataset is a tuple (e.g. for images there may be the image and a label), 
            this index describes which of the tuple elements enters into the resulting numpy array.

    Returns:
        nparray: numpy array containing the data
    """
    if isinstance(pt_data[0], tuple):
        result = np.empty((len(pt_data), ) +
                          tuple(pt_data[0][tuple_index].shape))
        for i in range(len(pt_data)):
            result[i] = pt_data[i][tuple_index]
            return result
    else:
        result = np.empty((len(pt_data), ) + tuple(pt_data[0].shape))
        for i in range(len(pt_data)):
            result[i] = pt_data[i]
        return result



import logging
from graphviz import Digraph

from numpy import load
from deepdiff import DeepDiff
from pailab.ml_repo.repo import MLObjectType, MLRepo, NamingConventions
from pailab.ml_repo.repo_objects import RepoInfoKey  # pylint: disable=E0401
import pailab.ml_repo.repo_store as repo_store
import pailab.ml_repo.repo_objects as repo_objects
logger = logging.getLogger(__name__)


class Node:
    def __init__(self, name):
        self.name = name
        self.depends_on = []
        self.modifies = []

    def add_dependence_on(self, node):
        if not node.name in self.depends_on:
            self.depends_on.append(node.name)

    def add_modifies(self, node):
        if not node.name in self.modifies:
            self.modifies.append(node.name)


class DependencyGraph:
    def __init__(self):
        self._nodes = {}

    def _add_node_if_not_exist(self, node):
        if not node in self._nodes:
            self._nodes[node] = Node(node)

    def add_dependency(self, node1, node2):
        self._add_node_if_not_exist(node1)
        self._add_node_if_not_exist(node2)
        self._nodes[node1].add_dependence_on(self._nodes[node2])
        self._nodes[node2].add_modifies(self._nodes[node1])

    def add_node(self, name):
        n = Node(name)
        self._nodes[name] = n

    def del_node(self, name):
        node = self._nodes[name]
        for n in node.modifies:
            self._nodes[n].depends_on.remove(name)
        for n in node.depends_on:
            self._nodes[n].modifies.remove(name)
        del self._nodes[name]


def get_dependency_graph(ml_repo, node='', category_to_color=None, ignore_categories=None):
    """Returns graph visualization from graphviz of th dependences around an object.

    This method creates a graph showing the dependencies between repo objects w.r.t. a special node or, if no node is specified, to all
    objects. To use this method, you must have installed graphviz.

    
    Args:
        ml_repo (MLRepo): repository
        node (str, optional): Name of object whose dependenies will be drawn. Defaults to '' which means that all dependencies are considered.
        category_to_color (map, optional): Map defining for each MLObject category the color used in the graph. Defaults to None which leads to a default color map.
        ignore_categories (iterable of M;ObjectType, optional): All MLObjectTypes defined over the iterable are ignored in the graph. Defaults to None.
       
    Returns:
        graphviz.Digraph: The resultign dependency structure.
    """
    def get_remaining_nodes(remaining_nodes, node, dependency_graph, modifiers=True, depends_on=True):
        remaining_nodes.append(node)
        tmp = dependency_graph._nodes[node]
        if modifiers:
            for n in tmp.modifies:
                if n not in remaining_nodes:
                    get_remaining_nodes(
                        remaining_nodes, n, dependency_graph, modifiers=True, depends_on=False)
        if depends_on:
            for n in tmp.depends_on:
                if n not in remaining_nodes:
                    get_remaining_nodes(
                        remaining_nodes, n, dependency_graph, modifiers=False, depends_on=True)

    def get_default_category_to_color():
        category_to_color = {}
        for k in MLObjectType:
            category_to_color[k] = 'white'
        category_to_color[MLObjectType.CALIBRATED_MODEL] = 'wheat'
        category_to_color[MLObjectType.MODEL] = 'wheat'
        category_to_color[MLObjectType.EVAL_DATA] = 'palegreen2'
        category_to_color[MLObjectType.TRAINING_DATA] = 'palegreen2'
        category_to_color[MLObjectType.TEST_DATA] = 'palegreen2'
        category_to_color[MLObjectType.RAW_DATA] = 'palegreen2'
        category_to_color[MLObjectType.TRAINING_FUNCTION] = 'lightgrey'
        category_to_color[MLObjectType.MODEL_EVAL_FUNCTION] = 'lightgrey'
        return category_to_color

    if ignore_categories is None:
        ignore_categories = [
            MLObjectType.LABEL, MLObjectType.COMMIT_INFO, MLObjectType.MAPPING, MLObjectType.JOB]
    dependency_graph = DependencyGraph()
    name_to_category = {}
    for k in MLObjectType:
        if k in ignore_categories:
            continue
        names = ml_repo.get_names(k.value)
        for n in names:
            name_to_category[n] = k
            obj = ml_repo.get(n)
            for l, v in obj.repo_info.modification_info.items():
                obj2 = ml_repo.get(l)
                dependency_graph.add_dependency(n, l)

    if node is not '':  # remove all nodes that are not connected to
        remaining_nodes = []
        all_nodes = [n for n in dependency_graph._nodes.keys()]
        get_remaining_nodes(remaining_nodes, node, dependency_graph)
        for k in all_nodes:
            if k not in remaining_nodes:
                dependency_graph.del_node(k)
    if category_to_color is None:
        category_to_color = get_default_category_to_color()

    dgr = Digraph(comment='dependency graph')
    for key in dependency_graph._nodes.keys():
        label_name = key+"\\n"+str(name_to_category[key].value)
        if key == node:
            dgr.node(key, label=label_name, color='red',
                     fillcolor='red', style='filled')
        else:
            dgr.node(key, label=label_name,
                     fillcolor=category_to_color[name_to_category[key]], style='filled')
    for key, node in dependency_graph._nodes.items():
        for d in node.modifies:
            dgr.edge(key, d)
    return dgr



# -*- coding: utf-8 -*-
"""Module defining classes to store numpy data in hdf5 files.

This module provides implementations of the :py:class:`pailab.ml_repo.repo_store.NumpyStore` using hdf5 file format.
"""
import h5py
import os
import pathlib
import logging
from pailab.ml_repo.repo_store import NumpyStore
logger = logging.getLogger(__name__)


def trace(aFunc):
    """ Trace entry, exit and exceptions.
    """
    def loggedFunc(*args, **kw):
        logger.info("Entering " + aFunc.__name__)
        try:
            result = aFunc(*args, **kw)
        except Exception as e:
            logger.exception("Error in " + aFunc.__name__ + ':' + str(e))
            raise
        logger.info("Exit " + aFunc.__name__)
        return result
    loggedFunc.__name__ = aFunc.__name__
    loggedFunc.__doc__ = aFunc.__doc__
    return loggedFunc


class NumpyHDFStorage(NumpyStore):
    """ Storage using hdf5 files to store numpy data.
    
    Example:
        Setup storage using folder ``C:\\temp\\data``::

            >>> store = NumpyHDFStorage('C:\\temp\\data')
    Args:
        folder (str): main directory where the files will be stored
        version_files (bool): If True, each version is contained in a separate file, otherwise all versions are in one file.
            If you like to work in a distributed environmnt (e.g. multiple users working in parallel) you should set this parameter to True so that no file merge is necessary.
            . Defaults to False.
    
    """

    def __init__(self, folder, version_files=False):
        self.main_dir = folder
        self._version_files = version_files
        if not os.path.exists(self.main_dir):
            os.makedirs(self.main_dir)

    def _create_file_name(self, name, version, change_if_not_exist=False):
        """ Function to create the file name for an object 

        Args:
            name (str): the identifier object
            version (str): the version id
            change_if_not_exist (bool): change the name if it does not exist. Defaults to False.

        Returns:
            str -- the filename
        """

        if self._version_files:
            filename = name + '_' + version + '.hdf5'
            if change_if_not_exist:
                if not os.path.exists(self.main_dir + '/' + filename):
                    return name + '.hdf5'
            return filename
        else:
            return name + '.hdf5'

    @trace
    def _delete(self, name, version):
        """ Delete an object with a predefined version

        Args:
            name (str): identifier of the object
            version (str): the version id
        """

        pass

    @staticmethod
    @trace
    def _save(data_grp, ref_grp, numpy_dict):
        """ saving the data 

        Args:
            data_grp ([type]): the data group
            ref_grp ([type]): the reference group 
            numpy_dict (numpy dict): the numpy dictionary to save

        Raises:
            NotImplementedException: [description]
        """

        for k, v in numpy_dict.items():
            if v is not None:
                if len(v.shape) == 1:
                    tmp = data_grp.create_dataset(
                        k, data=v, maxshape=(None, ))
                    ref_grp.create_dataset(k, data=tmp.regionref[0:v.shape[0]])
                else:
                    if len(v.shape) == 2:
                        tmp = data_grp.create_dataset(
                            k, data=v, maxshape=(None, v.shape[1]))
                        ref_grp.create_dataset(
                            k, data=tmp.regionref[0:v.shape[0], 0:v.shape[1]])
                    else:
                        if len(v.shape) == 3:
                            tmp = data_grp.create_dataset(
                                k, data=v, maxshape=(None, v.shape[1], v.shape[2]))
                            ref_grp.create_dataset(
                                k, data=tmp.regionref[0:v.shape[0], 0:v.shape[1], 0:v.shape[2]])
                        else:
                            if len(v.shape) == 4:
                                tmp = data_grp.create_dataset(
                                    k, data=v, maxshape=(None, v.shape[1], v.shape[2], v.shape[3]))
                                ref_grp.create_dataset(
                                    k, data=tmp.regionref[0:v.shape[0], 0:v.shape[1], 0:v.shape[2], 0:v.shape[3]])
                            else:                                        
                                raise NotImplementedException(
                                    'Not implemenet for dim>3.')

    @trace
    def add(self, name, version, numpy_dict):
        """ Add numpy data from an object to the storage.

        Args:
            name (str): the identifier of the object to add
            version (str): the object version 
            numpy_dict (numpy dict): the numpy dictionary to add
        """

        tmp = pathlib.Path(self.main_dir + '/' + name + 'hdf')
        save_dir = tmp.parent
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        with h5py.File(self.main_dir + '/' + self._create_file_name(name, version), 'a') as f:
            grp_name = '/data/' + version + '/'
            logger.debug('Saving data ' + name +
                         ' in hdf5 to group ' + grp_name)
            grp = f.create_group(grp_name)
            ref_grp = f.create_group('/ref/' + str(version) + '/')
            NumpyHDFStorage._save(grp, ref_grp, numpy_dict)

    @trace
    def _append_same_file(self, name, version_old, version_new, numpy_dict):
        """ Append data to the same file

        Args:
            name (str): the object identifier
            version_old (str): the previous object version
            version_new (str): the next object version
            numpy_dict (numpy dict): the data to add as a numpy dictionary
        """

        with h5py.File(self.main_dir + '/' + self._create_file_name(name, version_new), 'a') as f:
            logger.debug('Appending data ' + name +
                         ' in hdf5 with version ' + str(version_new))

            try:
                ref_grp = f['/ref/' + str(version_new) + '/']
                grp = f['/data/' + str(version_new) + '/']
            except:
                ref_grp = f.create_group('/ref/' + str(version_new) + '/')
                grp = f.create_group('/data/' + str(version_new) + '/')

            grp_previous = f['/data/' + str(version_old) + '/']
            for k, v in numpy_dict.items():
                data = grp_previous[k]
                old_size = len(data)
                new_shape = [x for x in v.shape]
                new_shape[0] += old_size
                new_shape = tuple(new_shape)
                data.resize(new_shape)
                grp[k] = h5py.SoftLink(data.name)
                if len(data.shape) == 1:
                    data[old_size:new_shape[0]] = v
                    ref_grp.create_dataset(
                        k, data=data.regionref[0:new_shape[0]])
                else:
                    if len(data.shape) == 2:
                        data[old_size:new_shape[0], :] = v
                        ref_grp.create_dataset(
                            k, data=data.regionref[0:new_shape[0], :])
                    else:
                        if len(data.shape) == 3:
                            data[old_size:new_shape[0], :, :] = v
                            ref_grp.create_dataset(
                                k, data=data.regionref[0:new_shape[0], :, :])
                        else:
                            if len(data.shape) == 4:
                                data[old_size:new_shape[0], :, :, :] = v
                                ref_grp.create_dataset(
                                    k, data=data.regionref[0:new_shape[0], :, :, :])

    def _append_different_file(self, name, version_old, version_new, numpy_dict):
        """ Append data to a different file

        Args:
            name (str): the object identifier
            version_old (str): the previous object version
            version_new (str): the next object version
            numpy_dict (numpy dict): the data to add as a numpy dictionary
        """

        # save data to append in separate file
        self.add(name + '_append', version_new, numpy_dict)
        # get shape
        grp_name_old = '/data/' + str(version_old) + '/'
        grp_name_new = '/data/' + str(version_new) + '/'
        old_filename = self._create_file_name(name, version_old)
        new_filename = self._create_file_name(name, version_new)
        tmp_filename = self._create_file_name(name + '_append', version_new)
        with h5py.File(self.main_dir + '/' + old_filename, 'r') as f_old:
            with h5py.File(self.main_dir + '/' + tmp_filename, 'r') as f_tmp:
                with h5py.File(self.main_dir + '/' + new_filename, 'w') as f_new:
                    ref_grp = f_new.create_group(
                        '/ref/' + str(version_new) + '/')
                    grp = f_new.create_group('/data/' + str(version_new) + '/')

                    for k, v in numpy_dict.items():
                        grp_old_k = f_old[grp_name_old][k]
                        grp_new_k = f_tmp[grp_name_new][k]
                        shape = (
                            grp_old_k.shape[0]+grp_new_k.shape[0], ) + grp_old_k.shape[1:]
                        layout = h5py.VirtualLayout(shape=shape)
                        layout[0:grp_old_k.shape[0]
                               ] = h5py.VirtualSource(grp_old_k)
                        layout[grp_old_k.shape[0]:] = h5py.VirtualSource(grp_new_k)
                        tmp = grp.create_virtual_dataset(k, layout)
                        ref_grp.create_dataset(k, data=tmp.regionref[:])
                        #ref_grp.create_virtual_dataset(k, layout)

    @trace
    def append(self, name, version_old, version_new, numpy_dict):
        """ append data to the an existing object

        Args:
            name (str): the object identifier
            version_old (str): the previous object version
            version_new (str): the next object version
            numpy_dict (numpy dict): the data to add as a numpy dictionary
        """

        if not self._version_files:
            self._append_same_file(name, version_old, version_new, numpy_dict)
        else:
            self._append_different_file(
                name, version_old, version_new, numpy_dict)

    @trace
    def get(self, name, version, from_index=0, to_index=None):
        """ get the numpy object for a name and a version, rows can be used

        Args:
            name (str): identifier of the object
            version (str): version of the object
            from_index (int): the index from which the data should be taken. Defaults to 0.
            to_index (int or None): the index to which the data is returned (None means till the end). Defaults to None.

        Raises:
            Exception: raises an exception if no object with the name exists
            Exception: raises an exception if no object and with the version exists 

        Returns:
            numpy array -- the numpy object to return
        """

        # \todo auch hier muss die referenz einschl. Namen verwendet werden
        with h5py.File(self.main_dir + '/' + self._create_file_name(name, version, change_if_not_exist=True), 'r') as f:
            grp_name = '/data/' + str(version) + '/'
            ref_grp = '/ref/' + str(version) + '/'
            logger.debug('Reading object ' + name +
                         ' from hdf5, group ' + grp_name)
            grp = f[grp_name]
            ref_g = f[ref_grp]
            result = {}
            for k, v in ref_g.items():
                result[k] = grp[k][v[()]]
        # todo needs improvement, handle indices in reading
        if from_index != 0 or (to_index is not None):
            tmp = {}
            if from_index != 0 and (to_index is not None):
                for k, v in result.items():
                    tmp[k] = v[from_index:to_index, :]
            else:
                if from_index != 0:
                    for k, v in result.items():
                        tmp[k] = v[from_index:-1, :]
                if to_index is not None:
                    for k, v in result.items():
                        tmp[k] = v[0:to_index, :]
            return tmp
        return result

    def object_exists(self, name, version):
        """ checks whether the object exists

        Args:
            name (str): the identifier of the object
            version (str): the version of the object

        Returns:
            bool -- returns true if the object exists
        """

        result = False
        try:
            with h5py.File(self.main_dir + '/' + self._create_file_name(name, version, change_if_not_exist=True), 'a') as f:
                grp_name = '/data/' + str(version) + '/'
                result = grp_name in f
        except:
            pass
        return result


def _get_all_files(directory):
    """ Returns set of all files in directory

    Args:
        directory (str): the directory

    Returns:
        set -- set with all filenames (including relative paths)
    """

    f = set()
    for path, subdirs, files in os.walk(directory):
        for name in files:
            p = path + '/' + name  # os.path.join(directory, name)
            p = p.replace(directory, '')
            #path.replace(directory, "") + name
            if p[0] == '\\' or p[0] == '/':
                p = p[1:]
            f.add(p)
    return f

import time
from contextlib import contextmanager

@contextmanager
def _lock_dir(main_dir, wait_time, timeout):
    _time = 0
    while _time < timeout:
        try:
            file = open(main_dir + '/.lock', 'x')
            file.close()
            break
        except:
            time.sleep(wait_time)
            _time += wait_time
    yield
    if _time >= timeout:
        raise Exception('Cannot obtain lock due to timeout. Either increase timeout or remove .lock in ' + main_dir + ' to make everything working.')
    if os.path.exists(main_dir + '/.lock'):
        os.remove(main_dir + '/.lock')

def _create_remote(remote_type, **kwargs):
    if remote_type == 'gcs':
        from pailab.ml_repo.remote_gcs import RemoteGCS
        return RemoteGCS(**kwargs)
    raise Exception('Unknown remote type ' + remote_type)


class NumpyHDFRemoteStorage(NumpyHDFStorage):
    """Storage working like NumpyHDFStorage locally but in addition provides synchronization with a remote.

    This storage stores numpy data in hdf5 files in a directory. It works very similar to the :py:class:`NumpyHDFStorage` with the difference
    that it synchronizes the data with a given remote (downloads and uploads the respective files).

    Example:
        This example shows how to setup the storage so that the data is stored in a 
        local directory and it can be synchronized ith googl ecloud storage::

            >>> numpy = NumpyHDFRemoteStorage('C:\\tmp\\data')
            >>> from pailab.ml_repo.remote_gcs import RemoteGCS
            >>> remote = RemoteGCS(bucket='my_data')
            >>> numpy.set_remote(remote)

    Args:
       folder (str): folder where data is stored
       remote_store (obj or dict): object representing a remote storage (e.g. :py:class:`pailab.ml_repo.remote_gcs.RemoteGCS` for the google cloud storage) or dictionary defining the remote params so that it can be created 
       sync_get (bool): If True, tries to download data automatically if it does not exist locally, otherwise it checks only locally
       sync_add (bool): If True, added data will be directly uploaded to the remote
    """

    def __init__(self, folder, remote_store = None,  sync_get = False, sync_add = False):
        super(NumpyHDFRemoteStorage, self).__init__(folder, version_files = True)
        if isinstance(remote_store, dict):
            self._remote_store = _create_remote(remote_store['type'], **remote_store['config'])
        else:
            self._remote_store = remote_store
        # timeout in sec (waiting for another get/pull/push from another process)
        self._timeout = 10
        # time in seconds to wait if another process is currently pushing/pulling/getting
        self._wait_time = 5
        self._sync_get = sync_get
        self._sync_add = sync_add

    def set_remote(self, remote_store):
        self._remote_store = remote_store

    def get(self, name, version, from_index=0, to_index=None):
        if not self._sync_get:
            return super(NumpyHDFRemoteStorage, self).get(name, version, from_index, to_index)

        result = None
        try:
            result = super(NumpyHDFRemoteStorage, self).get(name, version, from_index, to_index)
        except:
            with _lock_dir(self.main_dir, self._wait_time, self._timeout):
                filename = self._create_file_name(name, version, change_if_not_exist=False)
                self._remote_store._download_file(self.main_dir + '/' + filename, filename)
                result = super(NumpyHDFRemoteStorage, self).get(name, version, from_index, to_index)
        return result

    def add(self, name, version, numpy_dict):
        super(NumpyHDFRemoteStorage, self).add(name, version, numpy_dict)
        if self._sync_add:
            filename = self._create_file_name(name, version)
            self._remote_store._upload_file( self.main_dir + '/' + filename, filename)

    def push(self):
        """ Push changes to an external repo.
        """

        with _lock_dir(self.main_dir, self._wait_time, self._timeout):
            remote_files = {x for x in self._remote_store._remote_file_list()}
            local_files = _get_all_files(self.main_dir)
            if '.lock' in local_files:
                local_files.remove('.lock')
            files_to_push = local_files-remote_files
            for f in files_to_push:
                self._remote_store._upload_file(self.main_dir + '/' + f, f)

    def pull(self):
        """ Pull changes from an external repo
        """
        with _lock_dir(self.main_dir, self._wait_time, self._timeout):
            remote_files = self._remote_store._remote_file_list()
            local_files = _get_all_files(self.main_dir)
            local_files.remove('.lock')
            files_to_pull = remote_files - local_files
            for f in files_to_pull:
                self._remote_store._download_file(self.main_dir + '/' + f, f)




"""Module for pailab to numpy

    This module defines all necessary objects and functions to use numpy for preprocessing within pailab
    Currently only preprocessing functions without a previous fit to the training data is supported.
"""

__version__ = '0.0.1'

import numpy as np

from pailab.ml_repo.repo_objects import repo_object_init
from pailab.ml_repo.repo_objects import RepoInfoKey
from pailab.ml_repo.repo_objects import Preprocessor
from pailab.ml_repo.repo import MLObjectType


def get_classname(x):
    return x.__class__.__module__ + '.' + x.__class__.__name__

# region Preprocessing


class NumpyPreprocessingParam:
    """Interfaces the parameters of the numpy preprocessing algorithms

    """

    @repo_object_init()
    def __init__(self, numpy_param):
        self.numpy_params = numpy_param

    def get_params(self):
        return self.numpy_params


class NumpyPreprocessor:
    """Class to store all numpy preprocessor
    """

    @repo_object_init()
    def __init__(self, preprocessor):
        self.preprocessor = preprocessor


def select_columns(preprocessor_param, data_x, column_names):

    selected_columns = preprocessor_param.get_params()['columns']
    # first find the indices of the columns to select

    idx = []
    for column in selected_columns:
        try:
            idx.append(column_names.index(column))
        except:
            raise Exception('Column ' + column +
                            ' was not found and can not be selected.')

    # select columns
    x_data = data_x[:, idx]
    x_coordinates = [column_names[i] for i in idx]

    return x_data, x_coordinates


def add_preprocessor_select_columns(repo, preprocessor_name='Numpy_Select_Columns', preprocessor_param={}):
    """Adds a new numpy preprocessor to a pailab MLRepo
    This is a specific preprocessor to select certain columns

    Args:
        repo ([type]): [description]
        preprocessor_name ([type], optional): Defaults to None. [description]
        preprocessor_param ([type], optional): Defaults to None. [description]
    """

    if not 'columns' in preprocessor_param:
        raise Exception(
            'The preprocessor parameters must have an entry columns')

    # add preprocessing
    repo.add_preprocessing_transforming_function(
        select_columns, repo_name=preprocessor_name + '/select_columns')

    numpy_param = NumpyPreprocessingParam(preprocessor_param,
                                          repo_info={RepoInfoKey.NAME.value: preprocessor_name + '/preprocessor_param',
                                                     RepoInfoKey.CATEGORY: MLObjectType.PREPROCESSOR_PARAM.value})

    repo.add(numpy_param, 'adding preprocessor parameter')

    repo.add_preprocessor(preprocessor_name, transforming_function=preprocessor_name+'/select_columns', fitting_function=None,
                          preprocessor_param=preprocessor_name + '/preprocessor_param')

# endregion



from setuptools import setup, find_packages

exec(open("pailab/version.py").read())


def readme():
    with open('README.rst') as f:
        return f.read()


install_requirements = [
    "numpy>=1.13",
    "plotly>=3.5.0",
    "h5py>=2.7",
    "pandas>=0.20",
    "nbformat",
    "deepdiff",
]


# Extra dependencies for externals
sklearn = [
    "sklearn",
]
tensorflow = [
    "tensorflow",
]

all_dependencies = sklearn + tensorflow

setup(name='pailab',
      version=__version__,
      description='pailab the integrated machine learning workbench',
      long_description=readme(),
      classifiers=[
          'Development Status :: 4 - Beta',
      ],
      keywords='platform ml version-control neural-network ml-workbench ai big data',
      url='https://github.com/pailabteam/pailab',
      author='pailabteam',
      author_email='pailab@rivacon.com',
      license='Apache 2.0',
      packages=find_packages(exclude=['doc', 'examples', 'tests']),
      install_requires=install_requirements,
          extras_require={
          'all': all_dependencies,
          'sklearn': sklearn,
          'tensorflow': tensorflow,
      },
      include_package_data=True,
      zip_safe=False)



"""This module creaes the data for plotting used by plot.py. The different methods return the respective data and information for plotting 
such as axes labels, titles, additional info which may be shown on hover in the plot encapsulated in a dictionary.
"""

import numpy as np
import logging
import warnings
from pailab.ml_repo.repo import MLObjectType, MLRepo, NamingConventions  # pylint: disable=E0401,E0611
from pailab.ml_repo.repo_objects import RepoInfoKey, RawData  # pylint: disable=E0401
from pailab.ml_repo.repo_store import RepoStore, LAST_VERSION, FIRST_VERSION, _time_from_version  # pylint: disable=E0401
from pailab.tools.interpretation import _get_MMD2_X_vs_abs_ptw_error_percentile

logger = logging.getLogger(__name__)


def _get_value_by_path(source, path):
    if not isinstance(path, list):
        p = path.split('/')
    else:
        p = path
    if len(p) == 0:
        return source
    if '[' in p[0]:
        tmp = p[0].split('[')
        name = tmp[0]
        index = int(tmp.split(']')[0])
        value = source[name][index]
    else:
        value = source[p[0]]
    if len(p) > 1:
        return _get_value_by_path(value, p[1:])
    return value


class _LabelChecker:
    """Checks wether a certain object and object version has a label
    """

    def __init__(self, ml_repo):
        self._labels = {}
        labels = ml_repo.get_names(MLObjectType.LABEL)
        for label in labels:
            l = ml_repo.get(label)
            if l.name in self._labels:
                self._labels[l.name][
                    l.version] = l.repo_info[RepoInfoKey.NAME]
            else:
                self._labels[l.name] = {
                    l.version: l.repo_info[RepoInfoKey.NAME]}

    def get_label(self, name, version):
        _name = name
        if len(_name.split('/')) == 1:
            _name = name + '/model'
        if name in self._labels.keys():
            if version in self._labels[name].keys():
                return self._labels[name][version]
        return None

    def get_model_and_version(self, label_name):
        """Returns a tuple of model and version if a label with label_name exists, otherwise it returns None.

        Args:
            label_name (str): Name of potential label
        """
        for model_name, v in self._labels.items():
            for version, label in v.items():
                if label == label_name:
                    return model_name, version
        return None


def get_measure_by_parameter(ml_repo, measure_names, param_name, data_versions=LAST_VERSION, training_param=False):
    """Returns for a (list of) measure(s) the measures and corresponding param values for a certain parameter 

    Args:
        ml_repo (MLRepo): the ml repo
        measure_names (str, list(str)): string or list of strings of measure names  
        param_name (str): name of parameter
        data_versions (version number, optional): Defaults to None. If not None, only values on measures on dta with this version number are used
    Returns:
        [dict]: dictionary of measure name to list of dictionaries containing the result, i.e. 
        model_version: version of model parameter
        param_version: version of the parameter of this data point
        param_name: the parameter value
        data_version: version of the underlying data
        train_data_version: version number of trainin data used to calibrate the model leading to this measure
        measure_version: version of measure
        value: measure value
    """
    label_checker = _LabelChecker(ml_repo)

    if isinstance(measure_names, str):
        measure_names = [measure_names]

    result_all = {}
    for measure_name in measure_names:
        data = str(NamingConventions.Data(NamingConventions.EvalData(
            NamingConventions.Measure(measure_name))))
        measures = ml_repo.get(measure_name, version=None,
                               modifier_versions={data: data_versions})
        if not isinstance(measures, list):
            measures = [measures]
        model_name = NamingConventions.CalibratedModel(
            NamingConventions.Measure(measure_name)
        )
        if training_param:
            p_name = str(NamingConventions.TrainingParam(model_name))
        else:
            p_name = str(NamingConventions.ModelParam(model_name))
        train_data = ml_repo.get_names(MLObjectType.TRAINING_DATA)[0]
        model_name = str(model_name)
        # eval_name

        result = []
        n_warnings = 1
        for x in measures:
            p = ml_repo.get(
                p_name, version=x.repo_info[RepoInfoKey.MODIFICATION_INFO][p_name])
            try:
                param_value = _get_value_by_path(p.get_params(), param_name)
                # get train data version
                model = ml_repo.get(
                    model_name, version=x.repo_info.modification_info[model_name])
                info = {'model_version': x.repo_info[RepoInfoKey.MODIFICATION_INFO][model_name],
                        param_name: param_value, 'param_version': p.repo_info[RepoInfoKey.VERSION],
                        'data_version': x.repo_info[RepoInfoKey.MODIFICATION_INFO][data],
                        'train_data_version': model.repo_info[RepoInfoKey.MODIFICATION_INFO][train_data],
                        'measure_version': x.repo_info[RepoInfoKey.VERSION], 'value': x.value}
                label = label_checker.get_label(
                    model_name, x.repo_info[RepoInfoKey.MODIFICATION_INFO][model_name])
                if label is not None:
                    info['model_label'] = label
                result.append(info)
            except:
                n_warnings += 1
                logger.warning('Could no retrieve parameter ' + p_name + ' for ' +
                               p.repo_info.name + ', version ' + p.repo_info.version)
        if n_warnings > 1:
            warnings.warn('There are ' + str(n_warnings) +
                          ' cases where the parameter could not be retrieved. See logging (logevel warning) for details.')
        result_all[measure_name] = result
    return result_all


def _get_obj_dict(ml_repo, objs, label_checker=None, category=None, _objs=None):
    if _objs is None:
        _objs = {}
    # first determine models (including their respective version) to be plotted

    def add_obj(obj_name, obj_version, objs_):
        if obj_name in objs_.keys():
            if isinstance(obj_version, list):
                objs_[obj_name].extend(obj_version)
            else:
                objs_[obj_name].append(obj_version)
        else:
            if isinstance(obj_version, list):
                objs_[obj_name] = obj_version
            else:
                objs_[obj_name] = [obj_version]

    if objs is None:
        if label_checker is not None:
            logging.info('No model specified, use all labeled models.')
            for k, v in label_checker._labels.items():
                for w in v.keys():
                    add_obj(k, w, _objs)
        if category is not None:
            if isinstance(category, list):
                names = []
                for n in category:
                    names.extend(ml_repo.get_names(n))
            else:
                names = ml_repo.get_names(category)
            for n in names:
                _get_obj_dict(ml_repo, n, category=category, _objs=_objs)
    elif isinstance(objs, tuple):
        add_obj(objs[0], objs[1], _objs)
    elif isinstance(objs, str):
        if label_checker is not None:
            tmp = label_checker.get_model_and_version(objs)
            if tmp is not None:
                add_obj(tmp[0], tmp[1], _objs)
            else:
                add_obj(objs, LAST_VERSION, _objs)
        else:
            add_obj(objs, LAST_VERSION, _objs)
    elif isinstance(objs, dict):
        # check if some entry does contain a None entry->then the entry defines a label and must be replaced by model and version
        _get_obj_dict(
            ml_repo, [(k, v,) for k, v in objs.items()], label_checker, category, _objs)
    else:
        for m in objs:
            _get_obj_dict(ml_repo, m, label_checker, category, _objs)
    return _objs


def get_pointwise_model_errors(ml_repo, models, data, coord_name=None, data_version=LAST_VERSION, x_coord_name=None, start_index=0, end_index=-1):
    """Compute pointwise errors for given models and data.

    The method plots histograms between predicted and real values of a certain target variable for reference data and models. 
    The reference data is described by the data name and the version of the data (as well as the targt variables name). The models can be described
    by 
    - a dictionary of model names to versions (a single version number, a range of versions or a list of versions)
    - just a model name (in this case the latest version is used)

    Args:
        ml_repo (MLRepo): [description]
        models (str or dict): A dictionary of model names to versions (a single version number, a range of versions or a list of versions) or 
                just a model name (in this case the latest version is used)

        data (str or list of str): Name of input data to be used for the error plot.
        coord_name (int or str, optional): Index or name of y-coordinate used for error measurement. If None, the first coordinate is used. Defaults to None.
        data_version (str, optional): Version of the input data used. Defaults to LAST_VERSION.
        x_coord_name (str): If specified it defines the respective x-coordinate that will additionally to the errors be returned. 
            If None, no x-values will be returned. Defaults to None.
    """
    label_checker = _LabelChecker(ml_repo)
    _data = data
    if isinstance(_data, str):
        _data = [data]

    #_models = get_model_dict(ml_repo, models)
    _models = _get_obj_dict(ml_repo, models, label_checker,
                            MLObjectType.CALIBRATED_MODEL)
    ref_data = ml_repo.get(_data[0], version=data_version, full_object=False)
    coord = 0
    if coord_name is None:
        coord_name = ref_data.y_coord_names[0]
    if isinstance(coord_name, int):
        coord_name = ref_data.y_coord_names[coord_name]
    coord = ref_data.y_coord_names.index(coord_name)
    result = {'title': 'pointwise error (' + coord_name + ')', 'data': {}}
    if x_coord_name is None:
        result['x0_name'] = 'model-target  [' + coord_name + ']'
    else:
        if isinstance(x_coord_name, int):
            x_coord_name = ref_data.x_coord_names[x_coord_name]
        result['x0_name'] = x_coord_name
        result['x1_name'] = 'model-target  [' + coord_name + ']'

    for d in _data:
        ref_data = ml_repo.get(d, version=data_version, full_object=True)

        for m_name, m_versions in _models.items():
            if len(m_versions) == 1:
                m_versions = m_versions[0]
            tmp = m_name.split('/')[0]
            eval_data_name = str(
                NamingConventions.EvalData(data=d, model=tmp))
            logging.info('Retrieving eval data for model ' + tmp + ', versions ' +
                         str(m_versions) + ' and data ' + d + ', versions ' + str(data_version))
            eval_data = ml_repo.get(
                eval_data_name, version=None, modifier_versions={m_name: m_versions, d: data_version}, full_object=True)
            if not isinstance(eval_data, list):
                eval_data = [eval_data]
            for eval_d in eval_data:
                error = ref_data.y_data[:, coord] - eval_d.x_data[:, coord]
                end = end_index
                if end > 0:
                    end = min(end, error.shape[0])
                tmp = {}
                if x_coord_name is None:
                    tmp['x0'] = error[start_index:end]
                else:
                    tmp['x1'] = error[start_index:end]
                    tmp['x0_name'] = x_coord_name
                    tmp['x0'] = ref_data.x_data[start_index:end,
                                                ref_data.x_coord_names.index(x_coord_name)]
                tmp['info'] = {d: str(data_version),
                               m_name: str(eval_d.repo_info[RepoInfoKey.MODIFICATION_INFO][m_name])}

                model_label = label_checker.get_label(
                    m_name, eval_d.repo_info[RepoInfoKey.MODIFICATION_INFO][m_name])
                if model_label is not None:
                    tmp['label'] = model_label

                result['data'][eval_data_name + ': ' +
                               str(eval_d.repo_info[RepoInfoKey.VERSION])] = tmp
    return result


def get_ptws_error_dist_mmd(ml_repo, model, data, x_coords=None, y_coords=None, start_index=0, end_index=-1, percentile=0.1,
                            cache=True, scale=True, metric='rbf',  **kwds):
    """Returns Squared Maximum Mean Distance (MMD) between the distributions of the x-data w.r.t. a percentile of the absolute pointwise
        errors along the y-coordinates.

    Args:
        ml_repo (MLRepo): [description]
        model (str or dict): A dictionary of model names (or labels) to versions (a single version number, a range of versions or a list of versions) or 
                just a model name (in this case the latest version is used)
        data (str or dict): A dictionary of data namesto versions (a single version number, a range of versions or a list of versions) or 
                just a data name (in this case the latest version is used)
        x_coords (int, str or list, optional): x-coordinate or list of x-coordinates used to comput the squared MMD. 
            If None, all x-coordinates are used. Defaults to None.
        y_coords (int str or list, optional): y-coordinate or list of y-coordinates used to comput the squared MMD. If None, all y-coordinates are used. Defaults to None.
        start_index (int, optional): Start index of data. Defaults to 0.
        end_index (int, optional): End index of data. Defaults to -1.
        percentile (float, optional): Percentile of absolute error defining the x-values. Defaults to 0.1.
        cache (bool, optional): If True, caching is used using the given MLRepo. Defaults to True.
        scale (bool, optional): If True, the x-cordntes will be scaled by sklearn StandardScaler. Defaults to True.
        metric (str or callable, optional): The metric to use when calculating kernel between instances in a feature array. defaults to 'rbf'.
            If metric is a string, it must be one of the metrics in sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS. 
            If metric is precomputed, X is assumed to be a kernel matrix. Alternatively, if metric is a callable function, 
            it is called on each pair of instances (rows) and the resulting value recorded. 
            The callable should take two arrays from X as input and return a value indicating the distance between them. 
            Currently, sklearn provides the following strings: ‘additive_chi2’, ‘chi2’, ‘linear’, ‘poly’, ‘polynomial’, ‘rbf’,
                                                ‘laplacian’, ‘sigmoid’, ‘cosine’ 
        **kwds: optional keyword parameters that are passed directly to the kernel function.

    Returns:
        list of dict: List of dictionary where each dictionary contains the squared MMD as well as the name 
        and version of underlying data and model and the x- and y-coordinates used.

    """
    label_checker = _LabelChecker(ml_repo)
    tmp = _get_obj_dict(ml_repo, model, label_checker,
                        MLObjectType.CALIBRATED_MODEL)
    _models = []
    for m, m_tmp in tmp.items():
        for m_v in m_tmp:
            tmp = ml_repo.get(m, version=m_v, throw_error_not_unique=False)
            if isinstance(tmp, list):
                _models.extend(tmp)
            else:
                _models.append(tmp)

    tmp = _get_obj_dict(ml_repo, data, None, [
                        MLObjectType.TRAINING_DATA, MLObjectType.TEST_DATA])
    _data = []
    for d, d_tmp in tmp.items():
        for d_v in d_tmp:
            tmp = ml_repo.get(d, version=d_v, throw_error_not_unique=False)
            if isinstance(tmp, list):
                _data.extend(tmp)
            else:
                _data.append(tmp)

    result = []
    if cache:
        cache_ = ml_repo
    else:
        cache_ = None
    # set coordinates
    #

    if x_coords is not None:
        if isinstance(x_coords, int) or isinstance(x_coords, str):
            x_coords = [x_coords]
        for i in range(len(x_coords)):
            if isinstance(x_coords[i], str):
                x_coords[i] = _data[0].x_coord_names.index(x_coords[i])

    if y_coords is not None:
        if isinstance(y_coords, int) or isinstance(y_coords, str):
            y_coords = [y_coords]
        for i in range(len(y_coords)):
            if isinstance(y_coords[i], str):
                y_coords[i] = _data[0].y_coord_names.index(y_coords[i])

    # loop over models and data
    #
    for m in _models:
        for d in _data:
            _eval_data_name = str(
                NamingConventions.EvalData(data=d.repo_info.name, model=m.repo_info.name.split('/')[0]))
            eval_data = ml_repo.get(_eval_data_name, None, modifier_versions={
                                    m.repo_info.name: m.repo_info.version}, full_object=True)
            tmp = ml_repo.get(d.repo_info.name,
                              version=d.repo_info.version, full_object=True)
            mmd = _get_MMD2_X_vs_abs_ptw_error_percentile(
                tmp, eval_data, x_coords, y_coords, cache=cache_, percentile=percentile, scale=scale, metric=metric, **kwds)
            if x_coords is not None:
                x_coord_names = [d.x_coord_names[i] for i in x_coords]
            else:
                x_coord_names = d.x_coord_names

            if y_coords is not None:
                y_coord_names = [d.y_coord_names[i] for i in y_coords]
            else:
                y_coord_names = d.y_coord_names

            for i in range(len(x_coord_names)):
                for j in range(len(y_coord_names)):
                    result.append({'model': m.repo_info.name, 'model version': m.repo_info.version, 'data': d.repo_info.name,
                                   'data version': d.repo_info.version, 'x-coord': x_coord_names[i], 'y-coord': y_coord_names[j], 'mmd': mmd[i, j]})
    return result


def get_data(ml_repo, data, x0_coord_name, x1_coord_name=None, start_index=0, end_index=-1):
    data_dict = _get_obj_dict(ml_repo, data, category=[
                              MLObjectType.TRAINING_DATA, MLObjectType.TEST_DATA])
    result = {'title': 'data distribution', 'data': {}}
    result['data'] = {}
    if x1_coord_name is not None:
        result['x1_name'] = x1_coord_name
    x_coord = None
    y_coord = None
    for k, v in data_dict.items():
        ref_data = ml_repo.get(k, version=v, full_object=True)
        if not isinstance(ref_data, list):
            ref_data = [ref_data]
        for d in ref_data:
            if x_coord is None:
                if isinstance(x0_coord_name, str):
                    x_coord = d.x_coord_names.index(x0_coord_name)
                else:
                    x_coord = x0_coord_name
                    x0_coord_name = d.x_coord_names[x0_coord_name]
            if x1_coord_name is not None:
                y_coord = d.x_coord_names.index(x1_coord_name)
            tmp = {'info': {k: v}}
            tmp['x0'] = d.x_data[start_index:end_index, x_coord]
            if y_coord is not None:
                tmp['x1'] = d.x_data[:, y_coord]
            result['data'][d.repo_info[RepoInfoKey.NAME] + ': ' +
                           str(d.repo_info[RepoInfoKey.VERSION])] = tmp
    result['x0_name'] = x0_coord_name
    return result


def get_measure_history(ml_repo, measure_names):
    """Returns for a (list of) measure(s) the historic evolution of the measure  (using the order induced by the datetime encoded in the version number)

    Args:
        ml_repo (MLRepo): the ml repo
        measure_names (str, list(str)): string or list of strings of measure names (inlcuding full path) 

    Returns:

    """
    label_checker = _LabelChecker(ml_repo)

    if isinstance(measure_names, str):
        measure_names = [measure_names]

    result_all = {}
    for measure_name in measure_names:
        data = str(NamingConventions.Data(NamingConventions.EvalData(
            NamingConventions.Measure(measure_name))))
        measures = ml_repo.get(measure_name, version=(
            RepoStore.FIRST_VERSION, RepoStore.LAST_VERSION))  # , modifier_versions={data: data_versions})
        if not isinstance(measures, list):
            measures = [measures]
        model_name = NamingConventions.CalibratedModel(
            NamingConventions.Measure(measure_name)
        )
        model_name = str(model_name)
        train_data = ml_repo.get_names(MLObjectType.TRAINING_DATA)[0]

        # eval_name

        result = []
        for x in measures:
            info = {'model_version': x.repo_info[RepoInfoKey.MODIFICATION_INFO][model_name],
                    'data_version': x.repo_info[RepoInfoKey.MODIFICATION_INFO][data],
                    'train_data_version': x.repo_info[RepoInfoKey.MODIFICATION_INFO][train_data],
                    'value': x.value, 'datetime': _time_from_version(x.repo_info[RepoInfoKey.MODIFICATION_INFO][model_name])}
            label = label_checker.get_label(
                model_name, x.repo_info[RepoInfoKey.MODIFICATION_INFO][model_name])
            if label is not None:
                info['model_label'] = label
            result.append(info)

        result_all[measure_name] = result
    return result_all


def project(ml_repo, model=None, labels=None, left=None, right=None,
            output_index=0, n_steps=100):
    if left is None or right is None:
        raise Exception('Please specify left and right points.')
    steps = [0.0 + float(x)/float(n_steps-1) for x in range(n_steps)]
    # compute input for evaluation
    shape = (len(steps),)
    shape += left.shape
    x_data = np.empty(shape=shape)
    for i in range(len(steps)):
        x_data[i] = (1.0-steps[i])*left + steps[i]*right

    result = {}
    models = []
    if labels is None:
        labels = ml_repo.get_names(MLObjectType.LABEL)
        for l in labels:
            tmp = ml_repo.get(l)
            models.append((tmp.name, tmp.version, l))
    if model is None:
        tmp = ml_repo.get_names(MLObjectType.CALIBRATED_MODEL)
        if len(tmp) == 1:
            m = ml_repo.get(tmp[0])
            models.append((tmp[0], m.repo_info.version, 'latest'))
    else:
        m = ml_repo.get(model)
        model.append((model, m.repo_info.version, 'latest'))

    data = RawData(x_data, [""]*x_data.shape[1])
    for model in models:
        tmp = ml_repo.get(model[0], version=model[1], full_object=True)
        model_name = model[0].split('/')[0]
        model_def = ml_repo.get(
            model_name, tmp.repo_info.modification_info[model_name])
        eval_func = ml_repo.get(model_def.eval_function)
        tmp = eval_func.create()(tmp, data)
        if len(tmp.shape) == 1:
            result[model[2]] = tmp
        elif len(tmp.shape) == 2:
            result[model[2]] = tmp[:, output_index]

    return result



from contextlib import closing
import traceback
import datetime
import os
import sqlite3
import time
import abc

import uuid
from enum import Enum
from pailab.ml_repo.repo_objects import repo_object_init, RawData, RepoInfoKey  # pylint: disable=E0401
from pailab.ml_repo.repo import MLObjectType  # pylint: disable=E0401

import logging
logger = logging.getLogger(__name__)
logger_sql = logging.getLogger(__name__ + '_SQLITE')


class JobState(Enum):
    """Job states
    """
    SUBMITTED = 'submitted'
    WAITING_PRED = 'waiting_for_predecessor'
    WAITING = 'waiting'
    RUNNING = 'running'
    SUCCESSFULLY_FINISHED = 'successfully_finished'
    FAILED = 'failed'


class JobInfo:
    def __init__(self, user):
        self.user = user
        self.state = JobState.SUBMITTED.value
        self.submission_time = None
        self.error_message = None
        self.trace_back = None
        self.start_time = None
        self.end_time = None

    def set_state(self, state):
        self.state = state.value

    def set_start_time(self):
        self.start_time = datetime.datetime.now()

    def set_end_time(self):
        self.end_time = datetime.datetime.now()

    def __str__(self):
        result = self.user + ', ' + self.state + ', started ' + \
            str(self.start_time) + ', finished ' + str(self.end_time)
        if self.error_message is not None:
            result = result + ',  error_message: ' + self.error_message
        return result


class JobRunnerBase(abc.ABC):
    """Baseclass for all job runners so that they can be used together with the MLRepo
    """

    @abc.abstractmethod
    def add(self, job_name, job_version, user):  # pragma: no cover
        """[summary]
        """
        pass

    @abc.abstractmethod
    def get_info(self, job_name, job_version):  # pragma: no cover
        """[summary]
        """
        pass


class SimpleJobRunner(JobRunnerBase):
    def __init__(self, repo, throw_job_error = False):
        """Constructor
        
        Args:
            :param: repo (MLRepo): repository 
            throw_job_error (bool, optional): Defaults to False. If True, all errors from job will be rethrown. Otherwise, 
                error and traceback are stored in job only and no error will be thrown.
        """

        self._repo = repo
        self._job_info = {}
        self._throw_job_error = throw_job_error

    def set_repo(self, repo):
        self._repo = repo

    def get_info(self, job_name, job_version):
        jobid = job_name + ':' + str(job_version)
        return self._job_info[jobid]

    def add(self, job_name, job_version, user):
        job_id = job_name + ':' + str(job_version)
        job_info = JobInfo(user)
        job_info.set_state(JobState.RUNNING)
        job_info.start_time = datetime.datetime.now()
        self._job_info[job_id] = job_info
        job = self._repo.get(job_name, version=job_version)
        try:
            job.run(self._repo, job_id)
            job_info.end_time = datetime.datetime.now()
            job_info.set_state(JobState.SUCCESSFULLY_FINISHED)
        except Exception as e:
            logger.error(str(e) + ': ' + str(traceback.format_exc()))
            job_info.end_time = datetime.datetime.now()
            job_info.set_state(JobState.FAILED)
            job_info.error_message = str(e)
            job_info.trace_back = traceback.format_exc()
            if self._throw_job_error:
                raise e

        return job_id

    def get_waiting_jobs(self):
        """Return list of open jobs

        Returns:
            empty list because by construction, this JobRunner can only return something if the jobs have been finished
        """
        return []


class SQLiteJobRunner(JobRunnerBase):
    # region private
    def _create_new_db(self):
        logger.info('Creating new database for job runner.')
        self._conn = sqlite3.connect(self._sqlite_db_name)
        with closing(self._conn.cursor()) as cursor:
            cursor.execute('''CREATE TABLE predecessors (job_name TEXT NOT NULL, job_version TEXT NOT NULL, predecessor_name TEXT NOT NULL, predecessor_version TEXT NOT NULL, PRIMARY KEY(job_name, job_version))''')
            cursor.execute(
                '''CREATE TABLE jobs (job_name TEXT NOT NULL, job_version TEXT NOT NULL, job_state TEXT NOT NULL, start_time TIMESTAMP,
                                            end_time TIMESTAMP, error_message TEXT, stack_trace TEXT, insert_time TIMESTAMP DEFAULT CURRENT_TIMESTAMP NOT NULL,
                                            unfinished_pred_jobs INTEGER, user TEXT NOT NULL, PRIMARY KEY(job_name, job_version))''')
            self._conn.commit()
        
    def _setup_new(self):
        if not os.path.exists(self._sqlite_db_name):
            self._create_new_db()
        else:
            self._conn = sqlite3.connect(self._sqlite_db_name)

    @staticmethod
    def sqlite_name(name):
        return "'" + name + "'"

    def _set_finished(self, job_name, job_version, error_message='', stack_trace=''):
        '''Set the job finished

            It sets the job to finished and removes predecessor conditions respectively
        Args:
            job_name (str): name of job
            job_version (str): version of job
            error_message (str, optional): Defaults to ''. Error message if job raised an error
            stack_trace (str, optional): Defaults to ''. Stack trace of error.
        '''
        successfull = True
        if error_message != '' or stack_trace != '':
            successfull = False
        # first update jobs waiting for the job to be finished
        with closing(self._conn.cursor()) as cursor:
            if successfull:
                candidates = []
                for row in cursor.execute("select job_name, job_version from predecessors where predecessor_name = '"
                                        + job_name + "'" + " and predecessor_version = '" + job_version + "'"):
                    candidates.append((row[0], row[1]))
                cursor.execute( "delete from predecessors where predecessor_name = '"
                            + job_name + "'" + " and predecessor_version = '" + job_version + "'")
                self._conn.commit()
                for c in candidates:
                    cursor.execute("update jobs SET unfinished_pred_jobs=unfinished_pred_jobs-1  where job_name='" +
                                c[0] + "' and job_version='" + c[1] + "'")
                    self._conn.commit()
                    cursor.execute("update jobs SET job_state='" + JobState.WAITING.value + "' where job_name='" +
                                c[0] + "' and job_version='" + c[1] + "' and unfinished_pred_jobs <= 0")
                    self._conn.commit()
                cursor.execute("update jobs SET job_state='" + JobState.SUCCESSFULLY_FINISHED.value + "', end_time='" + str(datetime.datetime.now())
                            + "' where job_name = '" + job_name + "' and job_version = '" + job_version + "'")
                self._conn.commit()
            else:
                cursor.execute("update jobs SET job_state='" + JobState.FAILED.value + "', error_message='" + error_message.replace("'", "") + "', stack_trace='" + stack_trace.replace("'", "") + "', end_time='"
                            + str(datetime.datetime.now()) + "' where job_name = '" +
                            job_name + "' and job_version = '" + str(job_version) + "'")
                self._conn.commit()
        

    def _run_job(self, job_name, job_version):
        job = self._repo.get(job_name, version=job_version)
        try:
            job.run(self._repo, 0)
        except Exception as e:
            logger.error(str(e) + ': ' + str(traceback.format_exc()))
            return str(e), traceback.format_exc()
        return '', ''

    # endregion

    def __init__(self, sqlite_db_name, repo, sleep = 1, steps_to_heartbeat = 30):
        '''Contructor

        Args:
            sqlite_db_name (str): filename of sqlite database used
        '''
        self._sqlite_db_name = sqlite_db_name
        self._setup_new()
        self._sleep = sleep  # time to wait in sec before new request for open jobs to db
        self._steps_to_heartbeat = steps_to_heartbeat
        self._repo = repo
        self._id = str(uuid.uuid1())
        self._conn.set_trace_callback(logger_sql.info)

    def set_repo(self, repo):
        self._repo = repo

    def add(self, job_name, job_version, user):
        with closing(self._conn.cursor()) as cursor:
            job = self._repo.get(job_name, version=job_version)
            predecessors = job.get_predecessor_jobs()
            for predecessor in predecessors:
                cursor.execute("insert into predecessors (job_name, job_version, predecessor_name, predecessor_version) VALUES ("
                            + SQLiteJobRunner.sqlite_name(job_name) +
                            ", "
                            + SQLiteJobRunner.sqlite_name(str(job_version)) +
                            ","
                            + SQLiteJobRunner.sqlite_name(predecessor[0]) +
                            ","
                            + SQLiteJobRunner.sqlite_name(str(predecessor[1])) +
                            ")")
                self._conn.commit()
            job_state = JobState.WAITING.value
            if len(predecessors) > 0:
                job_state = JobState.WAITING_PRED.value
            cursor.execute("insert into jobs ( job_name, job_version, job_state,  unfinished_pred_jobs, user ) VALUES ("
                        + SQLiteJobRunner.sqlite_name(job.repo_info[RepoInfoKey.NAME]) +
                        ", "
                        + SQLiteJobRunner.sqlite_name(str(job.repo_info[RepoInfoKey.VERSION])) +
                        ",'"
                        + job_state +
                        "',"
                        + str(len(predecessors)) +
                        ", '" + user + "')")
            self._conn.commit()

    def run(self, max_steps=None):
        wait = self._sleep
        step = 0
        with closing(self._conn.cursor()) as cursor: 
            while True:
                if max_steps is not None:
                    step += 1
                    if step > max_steps:
                        return
                if wait > self._steps_to_heartbeat:
                    logger.info('heartbeat')
                    wait = 0
                row = None
                with self._conn:
                    cursor.execute("select job_name, job_version from jobs where job_state = '"
                                        + JobState.WAITING.value + "' order by insert_time desc")
                    row = cursor.fetchone()
                    if row is not None:
                        cursor.execute("update jobs SET start_time = '" + str(
                            datetime.datetime.now()) + "', job_state='" + JobState.RUNNING.value + "' where job_name = '"
                            + row[0] + "' and job_version='" + row[1] + "'")
                        self._conn.commit()
                    
                #logger.error('len(rows): ' + str(len(rows)))
                if row is not None:
                    logger.info('Start running job ' +
                                row[0] + ', version ' + row[1])
                    error, stack_trace = self._run_job(row[0], row[1])
                    self._set_finished(row[0], row[1], error, stack_trace)
                    if error == '' and stack_trace == '':
                        logger.info('Finished running job ' +
                                    row[0] + ', version ' + row[1] + ' successfully.')
                    else:
                        logger.error('Finished running job ' + row[0] + ', version ' +
                                    row[1] + ' with errors: ' + error + '   stacktrace: ' + stack_trace)
                    wait = 0
                else:
                    time.sleep(self._sleep)
                    wait += self._sleep
        
    def get_info(self, job_name, job_version):
        result = {}
        with closing(self._conn.cursor()) as cursor:
            rows = cursor.execute("select * from jobs where job_name='" +
                                job_name + "' and job_version = '" + str(job_version) + "'")
            column_names = [x[0] for x in rows.description]
            for row in rows:
                for i in range(len(row)):
                    result[column_names[i]] = row[i]
                cursor.close()
                return result
        return {'message': 'no info available for ' + job_name + ', version ' + str(job_version)}

    def get_waiting_jobs(self):
        """Return list of open jobs

        Returns:
            list of tuples: list containing tuples of job names and versions of the jobs currently waiting
        """
        open_jobs = []
        with closing(self._conn.cursor()) as cursor:
            for row in cursor.execute( "select job_name, job_version from jobs where job_state in ('"
                                    + JobState.WAITING.value + "','" + JobState.WAITING_PRED.value + "','" + JobState.RUNNING.value +"')"):
                open_jobs.append((row[0], row[1]))
        return open_jobs

    def close_connection(self):
        """Closes the database connection
        """
        self._conn.close()



"""Module for pailab to sklearn

    This module defines all necessary objects and functions to use sklearn from within pailab.
"""

__version__ = '0.0.1'

from pailab.ml_repo.repo_objects import repo_object_init
from pailab.ml_repo.repo_objects import RepoInfoKey
from pailab.ml_repo.repo_objects import Preprocessor
from pailab.ml_repo.repo import MLObjectType


def get_classname(x):
    return x.__class__.__module__ + '.' + x.__class__.__name__


class SKLearnModelParam:
    """Interfaces the parameters of the sklearn algorithms

    """

    @repo_object_init()
    def __init__(self, model, sklearn_param):
        self.sklearn_module_name = model.__class__.__module__
        self.sklearn_class_name = model.__class__.__name__
        self.sklearn_params = sklearn_param

    def get_params(self):
        return self.sklearn_params


class SKLearnModel:
    """Class to store all sklearn models in pailab's MLRepo
    """

    @repo_object_init()
    def __init__(self, model, preprocessors=None):
        self.model = model
        self.preprocessors = preprocessors


def eval_sklearn(model, data):
    """Function to evaluate an sklearn model

    Args:
        model ([type]): [description]
        data ([type]): [description]

    Returns:
        [type]: [description]
    """

    return model.model.predict(data.x_data)


def train_sklearn(model_param, data, preprocessors=None):
    def get_class(full_class_name):
        parts = full_class_name.split('.')
        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)
        return m
    m = get_class(model_param.sklearn_module_name +
                  '.' + model_param.sklearn_class_name)
    data_x = data.x_data
    data_y = data.y_data
    model = m(**model_param.sklearn_params)
    model.fit(data_x, data_y)
    result = SKLearnModel(model, preprocessors=preprocessors, repo_info={})
    return result


def add_model(repo, skl_learner, model_name=None, model_param=None,
              preprocessors=None):
    """Adds a new sklearn model to a pailab MLRepo

    Args:
        repo ([type]): [description]
        skl_learner ([type]): [description]
        model_name ([type], optional): Defaults to None. [description]
        model_param ([type], optional): Defaults to None. [description]
        preprocessors (list of strings, optional): List of used preprocessors
    """

    # model name
    m_name = model_name
    if m_name is None:
        m_name = skl_learner.__class__.__name__

    # check whether the preprocessors have been added

    # adding model functions
    repo.add_eval_function(eval_sklearn, repo_name='eval_sklearn')
    repo.add_training_function(train_sklearn, repo_name='train_sklearn')

    param = skl_learner.get_params(True)
    if model_param is not None:
        for k, v in model_param.items():
            param[k] = v
    model_p = SKLearnModelParam(
        skl_learner, param, repo_info={RepoInfoKey.NAME.value: m_name + '/model_param',
                                       RepoInfoKey.CATEGORY: MLObjectType.MODEL_PARAM.value})

    repo.add(model_p, 'adding model and training parameter')

    repo.add_model(m_name, model_eval='eval_sklearn', model_training='train_sklearn',
                   model_param=m_name + '/model_param', training_param='',
                   preprocessors=preprocessors)

# region Preprocessing


class SKLearnPreprocessingParam:
    """Interfaces the parameters of the sklearn algorithms

    """

    @repo_object_init()
    def __init__(self, preprocessor, sklearn_param):
        self.sklearn_module_name = preprocessor.__class__.__module__
        self.sklearn_class_name = preprocessor.__class__.__name__
        self.sklearn_params = sklearn_param

    def get_params(self):
        return self.sklearn_params


class SKLearnPreprocessor:
    """Class to store all sklearn preprocessor
    """

    @repo_object_init()
    def __init__(self, preprocessor):
        self.preprocessor = preprocessor


def transform_sklearn(preprocessor_param, data_x, x_coord_names, fitted_preprocessor=None):
    if fitted_preprocessor is None:
        def get_class(full_class_name):
            parts = full_class_name.split('.')
            module = ".".join(parts[:-1])
            m = __import__(module)
            for comp in parts[1:]:
                m = getattr(m, comp)
            return m
        m = get_class(preprocessor_param.sklearn_module_name +
                      '.' + preprocessor_param.sklearn_class_name)

        prepro = m(**preprocessor_param.sklearn_params)
    else:
        prepro = fitted_preprocessor
    return prepro.preprocessor.transform(X=data_x), x_coord_names


def fit_sklearn(preprocessor_param, data_x, x_coord_names):
    def get_class(full_class_name):
        parts = full_class_name.split('.')
        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)
        return m
    m = get_class(preprocessor_param.sklearn_module_name +
                  '.' + preprocessor_param.sklearn_class_name)

    if preprocessor_param.sklearn_params is not None:
        prepro = m(**preprocessor_param.sklearn_params)
    else:
        prepro = m()
    prepro.fit(X=data_x)
    return SKLearnPreprocessor(prepro, repo_info={})


def add_preprocessor(repo, skl_preprocessor, preprocessor_name=None, preprocessor_param=None):
    """Adds a new sklearn preprocessor to a pailab MLRepo

    Args:
        repo ([type]): [description]
        preprocessor (Class): The sklearn preprocessor class
        preprocessor_name ([type], optional): Defaults to None. [description]
        preprocessor_param ([type], optional): Defaults to None. [description]
    """

    # preprocessor name
    p_name = preprocessor_name
    if p_name is None:
        p_name = skl_preprocessor.__class__.__name__

    # add preprocessing
    repo.add_preprocessing_transforming_function(
        transform_sklearn, repo_name=p_name + '/transform_sklearn')
    repo.add_preprocessing_fitting_function(
        fit_sklearn, repo_name=p_name + '/fit_sklearn')
    param = skl_preprocessor.get_params(True)
    if preprocessor_param is not None:
        for k, v in preprocessor_param.items():
            param[k] = v
    skl_param = SKLearnPreprocessingParam(skl_preprocessor, param,
                                          repo_info={RepoInfoKey.NAME.value: p_name + '/preprocessor_param',
                                                     RepoInfoKey.CATEGORY: MLObjectType.PREPROCESSOR_PARAM.value})

    if skl_param is not None:
        repo.add(skl_param, 'adding preprocessor parameter')

    repo.add_preprocessor(p_name, transforming_function=p_name+'/transform_sklearn', fitting_function=p_name+'/fit_sklearn',
                          preprocessor_param=p_name + '/preprocessor_param')

# endregion



import logging
from pailab.ml_repo.repo import MLRepo, MLObjectType, NamingConventions
from pailab.ml_repo.repo_objects import RepoInfoKey  # pylint: disable=E0401
from pailab.ml_repo.repo_store import RepoStore  # pylint: disable=E0401
logger = logging.getLogger(__name__)




def get_initial_config(repo):
    model_names = repo.get_names(MLObjectType.MODEL)
    model_checks ={}
    for m in model_names:
        model_checks[m] = {'correct': False, 'check_for_latest': True, 'model_version': RepoStore.LAST_VERSION}
    
    model_checks = {'models': model_checks, 'labels': {'__ALL__': {'correct': False}}}

    data_checks = {'overlapping': True, 'usage': False}
    repo_checks = {'integrity': True}
    return {'model_checks': model_checks, 'data_checks': data_checks, 'repo_checks': repo_checks}

class Model:

    @staticmethod
    def __check_model(repo, model_name, correct, model_version, check_for_latest=False):
        """Check if the model is calibrated and evaluated on the latest versions

        Args:
            repo (MLRepo): the ml repo
            model_name (str): model name
            model_version (version): model version to check
            correct (bool): determine whether training, evaluations, measures and tests will be qutomatically triggered if check fails to correct. Defaults to False.
            check_for_latest (bool): if true, some additional checks are performed to see whether the latest model is calibrated on th latest data

        Returns:
            dict -- dictionary containing the modifier versions and the latest version of the objects, if dictionary is empty noc model inconsistencies could be found
        """
        logging.info('Checking model ' + model_name + ', version: ' + str(model_version) +
                        ', correct: ' + str(correct) + ', check_for_latest: ' + str(check_for_latest))
        model_main_name = model_name.split('/')[0]
        result = {}
        repo_store = repo.get_ml_repo_store()

        # first check if all versions of the models modifiers are still the latest version
        if check_for_latest:
            tmp = {}
            m = repo.get(MLRepo.get_calibrated_model_name(
                model_name), version=model_version)
            for k, v in m.repo_info[RepoInfoKey.MODIFICATION_INFO].items():
                latest_version = repo_store.get_latest_version(k)
                if not str(v) == str(latest_version):
                    tmp[k] = {'modifier version': v,
                                'latest version': latest_version}
            if len(tmp) > 0:
                result['latest model version not on latest inputs'] = tmp
                if correct == True:
                    job_id = repo.run_training(model_name)
                    job_ids = repo.run_evaluation(
                        model_name, message='running evaluations triggered by new training of ' + model_name, predecessors=[job_id])
                    job_ids = repo.run_measures(
                        model_name, message='running measures triggered by new training of ' + model_name, predecessors=job_ids)
                    job_ids = repo.run_tests(
                        model_name, message='running tests triggered by new training of ' + model_name, predecessors=job_ids)
                    return result  # we can return here since everything was corrected
        # simply check if model has been evaluated on the latest data sets
        data_names = repo.get_names(MLObjectType.TRAINING_DATA)
        data_names.extend(repo.get_names(MLObjectType.TEST_DATA))
        eval_check_result = {}
        data_corrected = set()
        for data in data_names:
            latest_data_version = repo_store.get_latest_version(data)
            eval_name = str(NamingConventions.EvalData(
                data=data, model=model_main_name))
            tmp = repo.get(eval_name, version=None, modifier_versions={
                model_name: model_version, data: latest_data_version}, 
                throw_error_not_exist=False, throw_error_not_unique=False)
            if tmp == []:
                eval_check_result[data] = latest_data_version
                if correct:
                    repo.run_evaluation(model_name, message='automatically corrected from _check_model',
                                        model_version=model_version, run_descendants=True)
                    data_corrected.add(data)

        if len(eval_check_result) > 0:
            result['evaluations missing'] = eval_check_result
        # simply check if all measures have been computed on latest data for the respective model
        measures = repo.get_names(MLObjectType.MEASURE_CONFIGURATION)
        measure_check_result = set()
        if len(measures) > 0:  # only if measures were defined
            for m in measures:
                measure_config = repo.get(m)
                for measure_name in measure_config.measures.keys():
                    for data in data_names:
                        tmp = str(NamingConventions.Measure(
                            str(NamingConventions.EvalData(data=data, model=model_main_name)), measure_type=measure_name))
                        obj = repo.get(tmp, version=None, modifier_versions={
                            model_name: model_version}, throw_error_not_exist=False, throw_error_not_unique=False)
                        if obj == []:
                            measure_check_result.add(tmp)
                            # we do only have to correct if the underlying evaluation data has not been corrected (then predecessors are included)
                            if correct and not data in data_corrected:
                                repo.run_measures(model_name, message='automatically corrected from _check_model',
                                                    model_version=model_version, run_descendants=True)

        if len(measure_check_result) > 0:
            result['measures not calculated'] = measure_check_result
        return result

    @staticmethod
    def run(repo: MLRepo, model_name=None, correct=False,
            model_version=RepoStore.LAST_VERSION, 
            model_label=None, check_for_latest = True):
        """Perform consistency checks for specified model versions

        Args:
            :param repo (MLRepo): ml repository
            :model_name (str, optional): Defaults to None. If specified, the model defined by the name and the model_version are checked.
            :param correct (bool, optional): Defaults to False. If True, the method starts the corresponding jobs to fix the found issues.
            :param model_version (str or list of str, optional): Defaults to RepoStore.LAST_VERSION. The model version(s) of the models to check
            :param model_label ([type], optional): Defaults to None. If it is set to '__ALL__', all labels are checked.

        Raises:
            Exception: Raises if a model version but no model name is specified 

        Returns:
            [dict]: dictionary mapping model+version to issues found. May be empty if no issues exist.
        """

        logger.info('Start checking model.')
        result = {}

        model_labels = []
        if model_label is not None:
            if isinstance(model_label, list):
                model_labels=model_label
            elif isinstance(model_label, str):
                if model_label == '__ALL__':
                    model_labels = repo.get_names(MLObjectType.LABEL)
                else:
                    model_labels=[model_label]

        for model_label in model_labels:  # check the model defined by the label
            label = repo.get(model_label)
            tmp = Model.__check_model(repo, label.name, correct,
                                    model_version=label.version,
                                    check_for_latest=False)
            if len(tmp) > 0:
                result[model_label] = tmp

        if model_name is not None:  # check the model defined by name and versions
            if len(model_name.split('/')) == 1:
                model_name = model_name + '/model'
            latest_version = repo.get(model_name).repo_info.version
            if model_version is None:
                model_version = RepoStore.LAST_VERSION
            if isinstance(model_version, str):
                model_versions = [model_version]
            else:
                model_versions = model_version
            for version in model_versions:
                if str(latest_version) == str(version) or str(version) == RepoStore.LAST_VERSION:
                    if str(version) == RepoStore.LAST_VERSION:
                        logger.debug(
                            'Latest version found, check if latest version ran on latest data.')
                        tmp = Model.__check_model(repo, model_name, correct,
                                                            latest_version, check_for_latest=check_for_latest)
                    else:
                        tmp = Model.__check_model(repo, model_name, correct,
                                                            version, check_for_latest=check_for_latest)
                else:
                    tmp = Model.__check_model(repo, model_name, correct,
                                        version, check_for_latest=check_for_latest)
                if len(tmp) > 0:
                    result[model_name+':' + str(version)] = tmp

        logger.info('Finished checking model.')
        return result

    
class Data:
    @staticmethod
    def __check_no_overlapping_training_test(repo):
        """Checks if current training data does overlap with test data

        Args:
            repo (MLRepo): ml repository
        """
        def compute_start_end_index(dataset, rawdata):
            start = dataset.start_index
            if start < 0:
                start = start + rawdata.n_data
            end = dataset.end_index
            if end is None:
                end = rawdata.n_data
            elif end < 0:
                end = rawdata.n_data + end
            return (start, end)

        def overlap(interval_1, interval_2):
            if interval_1[0] < interval_2[0] and interval_1[1] < interval_2[0]:
                return False
            if interval_2[0] < interval_1[0] and interval_2[1] < interval_1[0]:
                return False
            return True

        tmp = repo.get_names(MLObjectType.TRAINING_DATA)
        if len(tmp) == 0:
            raise Exception('Repo contains no training data.')
        training_data = repo.get(tmp[0])
        training_raw = repo.get(training_data.raw_data,
                                version=training_data.raw_data_version)
        training_indices = compute_start_end_index(
            training_data, training_raw)

        test_data_names = repo.get_names(MLObjectType.TEST_DATA)
        result = {}
        for data in test_data_names:
            test_data = repo.get(data)
            if test_data.raw_data == training_data.raw_data:
                test_data_raw = repo.get(
                    test_data.raw_data, version=test_data.raw_data_version)
                test_indices = compute_start_end_index(
                    test_data, test_data_raw)
                if overlap(training_indices, test_indices):
                    result[data] = {'training and test data overlap': {
                        data: test_data.repo_info.version, training_data.repo_info.name: training_data.repo_info.version}}
        return result

    @staticmethod
    def __check_usage(repo):
        """Check if all RawData is used

        Args:
            repo ([type]): [description]

        Raises:
            NotImplementedError: [description]

        Returns:
            [type]: [description]
        """

        def compute_start_end_index(dataset, rawdata):
            start = dataset.start_index
            if start < 0:
                start = start + rawdata.n_data
            end = dataset.end_index
            if end is None:
                end = rawdata.n_data
            elif end < 0:
                end = rawdata.n_data + end
            return (start, end)
        raise NotImplementedError()

    @staticmethod
    def run(repo, overlapping = True, usage = False):
        result = {}
        if overlapping:
            result = Data.__check_no_overlapping_training_test(repo)
        if usage:
            #todo must be implemented
            result.update(Data.__check_usage(repo))
        return result

class Repo:
    @staticmethod
    def __integrity(repo):
        ml_repo = repo.get_ml_repo_store()
        if hasattr(ml_repo, 'check_integrity'):
            return ml_repo.check_integrity()
        return {}

    @staticmethod
    def run(repo, integrity = True):
        if integrity:
            return Repo.__integrity(repo)

class Tests:
    @staticmethod
    def __check_test(repo: MLRepo, model, model_version, test_definition):
        
        # loop over all data
        data = test_definition._get_data(repo)
        results = {}
        for d in data:    
            # first create from definition the test for the given model to get the test name
            test_name = str(NamingConventions.Test(model=NamingConventions.get_model_from_name(model), test_name= test_definition.repo_info.name, data=d))
            logging.debug('Checking test ' + test_name + ' for ' + model +  ', version '+ model_version)
            test = repo.get(test_name, version = None, modifier_versions={model: model_version}, throw_error_not_exist=False, throw_error_not_unique=False)
            if test == []: 
                results[test_name] = 'Test for model ' + model + ', version ' + model_version +' on latest data ' + d + ' missing.'
                continue             
            if isinstance(test, list): # search latest test 
                t = test[0]
                for k in range(1, len(test)):
                    if test[k].repo_info.commit_date > t.repo_info.commit_date:
                        t = test[k] 
                test = t
            result = test._check(repo)
            if not result is None:
                results[test_name] = result
            if not test.result =='succeeded':
                results[test_name] = 'Test for model ' + model + ', version ' + model_version +' on latest data ' + d + ' failed, details: ' + str(test.details)
                
        return results

    @staticmethod
    def run(repo: MLRepo):
        test_definitions = repo.get_names(MLObjectType.TEST_DEFINITION)
        results = {}
        version_to_label = {}
        labels = repo.get_names(MLObjectType.LABEL)
        for l in labels:
            tmp = repo.get(l)
            version_to_label[tmp.version] = tmp.repo_info.name
        for t in test_definitions:
            test_definition = repo.get(t)
            models =   test_definition._get_models(repo)
            for m, v in models.items():
                for version in v:
                    result = Tests.__check_test(repo, m, version, test_definition)
                    if len(result) > 0:
                        if version in version_to_label:
                            results[m+':'+version_to_label[version]] = result
                        else:    
                            results[m+':'+version] = result
        return results

def run(repo: MLRepo, config : dict = None):
    
    if config is None:
        if 'check' not in repo._config.keys():
            config = get_initial_config(repo)
            repo._config['check'] = config
            repo._save_config()    
        config = repo._config['check']
    else:
        config = get_initial_config(repo)
        repo._config['check'] = config
        repo._save_config()

    result = []
    for k,v in config['model_checks']['models'].items():
        tmp = Model.run(repo, model_name=k, correct=v['correct'], model_version = v['model_version'], check_for_latest=v['check_for_latest'])
        if len(tmp) > 0:
            result.append(tmp)
    for k,v in config['model_checks']['labels'].items():
        tmp = Model.run(repo, correct=v['correct'], model_label = k, check_for_latest=False)
        if len(tmp) > 0:
            result.append(tmp)
    tmp = Data.run(repo, **config['data_checks'])        
    if len(tmp) > 0:
        result.append(tmp)
    tmp = Repo.run(repo, **config['repo_checks'])
    return result